; CPC/IP configuration
; Copyright (c) 1999-2001 Mark RISON
FALSE		EQU  0
TRUE		EQU  1
; Define the memory configuration
MEMCONFIG	EQU  64		; Must be 64 or 128 (K)
; Define the serial interface type and the serial parameters
AMSTRAD		EQU  TRUE
PACE		EQU  FALSE
CPCAI		EQU  FALSE
BIT_RATE	EQU  9600	; Must be divisible by 100
PARITY_NONE	EQU  TRUE
PARITY_ODD	EQU  FALSE
PARITY_EVEN	EQU  FALSE
STOP_1		EQU  TRUE
STOP_1_5	EQU  FALSE
STOP_2		EQU  FALSE
INTDRIVEN	EQU  TRUE
HWFLOWCONTROL	EQU  FALSE
; Define the data-link layer protocol
SLIP		EQU  TRUE
PPP		EQU  FALSE
; Define which bits should be included
PING		EQU  TRUE	; Ping client
IFCONFIG	EQU  TRUE	; IfConfig module
TIMEOUT		EQU  TRUE	; Timeout module
LOOPBACK	EQU  TRUE	; Loopback driver
HOSTS		EQU  TRUE	; hosts.txt module
UDP		EQU  TRUE	; UDP layer
TFTPD		EQU  TRUE	; TFTP daemon
DNSD		EQU  TRUE	; DNS server
RESOLV		EQU  TRUE	; Resolver module
HOST		EQU  TRUE	; DNS client
TCP		EQU  TRUE	; TCP layer
HTTPD		EQU  TRUE	; HTTP daemon
TELNET		EQU  TRUE	; Telnet client
FINGER		EQU  TRUE	; Finger client
; Debug options
SHOW_RAW_IN	EQU  FALSE
SHOW_RAW_OUT	EQU  FALSE
SHOW_INFO	EQU  FALSE
SHOW_ERROR	EQU  TRUE
SHOW_PPP	EQU  TRUE
DROP_OUT	EQU  FALSE	; Drop 1 in n output datalink packets (0 = don't, 1 = invalid)
MARK_HACKS	EQU  FALSE
; Protocol options
CHECKSUM_PPP_IN	EQU  TRUE	; Optional (but a good idea to keep LCP happy!)
CHECKSUM_IP_IN	EQU  TRUE	; Must be set for RFC1122-compliance (3.2.1.2)
CHECKSUM_ICMP_IN	EQU  TRUE	; RFC1122 doesn't seem to specify this
CHECKSUM_TFTP_OUT	EQU  TRUE	; For ACK packets (RFC112? FIXME)
DEST_UNREACH	EQU  TRUE	; RFC1122-optional (3.2.2.1)
; ASCII control codes
NUL_		EQU  0
BEL		EQU  7
BS		EQU  8
TAB		EQU  9
LF		EQU  10
CR		EQU  13
ESC		EQU  27
DEL		EQU  127
; Miscellaneous
PROMPT		EQU  ">"
SWITCH		EQU  "-"
; Symbol derivations
	IF  MEMCONFIG-64
IS64K		EQU  FALSE
IS128K		EQU  TRUE
	ELSE 
IS64K		EQU  TRUE
IS128K		EQU  FALSE
	ENDC 
	IF  INTDRIVEN
POLLED		EQU  FALSE
	ELSE 
POLLED		EQU  TRUE
	ENDC 
; Memory map for 128K memory configuration:
;
;	    -27ff	Free
;	2800-2fff	AMSDOS read buffer
;	3000-37ff	AMSDOS write buffer
;	3800-7cff	Code
;	7d00-7dff	Buffer for keyboard input
;	7e00-7eff	Reserved for mode 2 text accelerator
;	7f00-7fff	Reserved for screen blanker
;	8000-83ff	Buffer for IP datagram (rx, and tx response)
;	8400-87ff	Buffer for IP datagram (timeout/async tx)
;	8800-8fff	Special code
;	9000-		Free
;
;	Bank 4		RX serial buffer
;	Bank 5		TX serial buffer
;	Bank 6		Loopback buffer
;	Bank 7		Free
;
; Note it's all a bit of a juggling act.
;
; In the 128K memory configuration, the serial device driver and
; loopback driver can't be in bank 1 because they switch this out,
; so they have to be below &4000 or above &7fff.
;
; The components which use events, however, can't be in bank 0
; because event blocks must be in the central 32K, so have to be above &3fff!
;
; A polled serial device driver, which needs to do both,
; therefore has to be placed in a special area in bank 2.
;
; It's a bit simpler in the 64K memory configuration; the only concern here
; is the one about components using events not being in bank 0.
;
; The memory map for the 64K memory configuration changes as follows:
;
;	8800-8bff	TX serial buffer
;	8c00-8fff	RX serial buffer
;	9000-93ff	Loopback buffer
;	9400-		Free
; AMSDOS buffers are 2K in size and may be anywhere in RAM
amsdosreadbuf	EQU  0x2800
amsdoswritebuf	EQU  0x3000
; The keyboard buffer is 256B in size (see main.z) and must be page-aligned
keybbuffer	EQU  0x7d00
; The rx IP buffer is 1K in size
; and its base must have b9-b0 and b10 clear (see slip.z/ppp.z)
ipbuffer	EQU  0x8000
; The tx IP buffer is 1K in size
; but may be anywhere in RAM
iptxbuffer	EQU  0x8400
	IF  IS128K
; The loopback and serial tx and rx buffers are 16K in size
; and their bases are all at &4000 (see serialdd.z and loopback.z)
serialtxbuffer	EQU  0x4000
serialtxbufsiz	EQU  0x4000
serialrxbuffer	EQU  0x4000
serialrxbufsiz	EQU  0x4000
lobuffer	EQU  0x4000
lobufsiz	EQU  0x4000
specialcode	EQU  0x8800
	ELSE 		; !IS128K
; The serial tx buffer must be 1K in size
; and its base must have b9-b0 and b10 clear (see serialdd.z)
serialtxbuffer	EQU  0x8800
serialtxbufsiz	EQU  0x400
; The serial rx buffer must be 1K in size
; and its base must have b9-b0 clear and b10 set (see serialdd.z)
serialrxbuffer	EQU  0x8c00
serialrxbufsiz	EQU  0x400
; The loopback buffer must be 1K in size
; and its base must have b9-b0 and b10 clear (see loopback.z)
lobuffer	EQU  0x9000
lobufsiz	EQU  0x400
	ENDC 		; IS128K
	ORG  0x3800
	jp  main
config_hostaddr:	 
; This host's address
	db  192, 168, 6, 128	; Private address; see RFC1918
config_hosts_txt:	 
; The address of the hosts.txt data
	dw  hosts_txt
config_nameserveraddr:	 
; The address of the nameserver to use
	db  127, 0, 0, 1	; This host; maybe it's running a dnsd
config_dns_resourcerecords:	 
; The address of the DNS server zone data
	dw  dns_resourcerecords
	IF  PPP
config_pap_auth_req_pkt:	 
; This is the information to use in PPP authentication using PAP
; Remember to guard this file and the binary if you make use of PAP!
config_par_pktsiz	EQU  config_pap_auth_req_pktend-config_pap_auth_req_pkt
	db  PAP_AUTH_REQ, 0x42
	db  config_par_pktsiz/256, config_par_pktsiz .and. 0xff
	db  config_par_password-config_par_username-1
config_par_username:	 
	db  "USERNAME"
	db  config_pap_auth_req_pktend-config_par_password
config_par_password:	 
	db  "PASSWORD"
config_pap_auth_req_pktend:	 
	ds  16		; Allow a bit more space for username/password
	ENDC 
config_domain:	 
; This is the domain to append to dotless hostname address resolver lookups
	db  "bufo-house.org" , 0
