; CPC/IP IP layer
; Copyright (c) 1999-2001 Mark RISON
IP_VERIHL	EQU  0
IP_TOS		EQU  1
IP_LEN1		EQU  2
IP_LEN0		EQU  3
IP_ID1		EQU  4
IP_ID0		EQU  5
IP_FRG1		EQU  6
IP_FRG0		EQU  7
IP_TTL		EQU  8
IP_PRT		EQU  9
IP_CSM1		EQU  10
IP_CSM0		EQU  11
IP_SRC3		EQU  12
IP_SRC2		EQU  13
IP_SRC1		EQU  14
IP_SRC0		EQU  15
IP_DST3		EQU  16
IP_DST2		EQU  17
IP_DST1		EQU  18
IP_DST0		EQU  19
IP_HDRSIZE	EQU  20
IP_ICMP		EQU  1		; Values for IP_PRT
IP_TCP		EQU  6
IP_UDP		EQU  17
IP_DF		EQU  0x40	; Values for IP_FRG1
IP_MF		EQU  0x20
IP_DEFAULT_VERIHL	EQU  0x45	; IPv4, no options
IP_DEFAULT_TTL	EQU  64		; See RFC1122 (3.2.1.7) and RFC1700
LOOPBACK_ADDR	EQU  127
ip_handle:	 
; Handle IP datagram from ipbuffer to (ipbufferinsert) - 1
	ld  hl, (ipbufferinsert)
	ld  (hl), 0	; Useful for checksums if odd number of octets
	ld  de, ipbuffer	; (always safe because ipbufferinsert <= ipbuffer + &3ff)
	or  a
	sbc  hl, de	; HL = # of IP octets received
	ret  z		; Always quietly dump zero-len
; Are we in fact dealing with IPv4?
	ld  a, (ipbuffer+IP_VERIHL)
	ld  d, a
	and  0xf0
	cp  0x40
	jp  nz, ip_notv4
; The minimal IPv4 header size is 20
	ld  a, d
	and  0x0f
	rla 
	rla 
	cp  IP_HDRSIZE
	jr  c, ip_headertoosmall
; The header can't occupy more than the whole datagram!
	inc  h
	dec  h
	jr  nz, ip_nottoosmall
	dec  a
	cp  l
	jp  nc, ip_headertoobig
ip_nottoosmall:	 
	ld  ix, ipbuffer	; IX = base of IP octets
; Fragmentation is not supported
; This makes this IP implementation not RFC1122-compliant (3.3.2)
; (Not handling options also does (3.2.1.8))
; COULD add option-handling
	ld  a, (ipbuffer+IP_FRG1)
	and  0x3f	; Fragment offset and MF bit should all be zero
	or  (ix+IP_FRG0)
	jp  nz, ip_fragmented
; Does the received datagram size match the declared size?
; (Assuming no fragmentation)
	ld  d, (ix+IP_LEN1)
	ld  e, (ix+IP_LEN0)
	ex  de, hl
	sbc  hl, de	; C clear before because of OR above
	ex  de, hl
	jp  nz, ip_sizemismatch
; Check the checksum
	IF  CHECKSUM_IP_IN
	call  cs_calcipchecksum
	jr  nz, ip_badchecksum
	ENDC 
; Quietly dump packet if sent to Class D or Class E address
	ld  a, (ipbuffer+IP_SRC3)
	or  0x1f	; Class D/E start 1110/1111
	inc  a
	ret  z
; FIXME should check correct destination for RFC1122-compliance (3.2.1.3)
; (perhaps just dump anything which isn't exactly our host address)
; It's looking good!
	IF  SHOW_INFO
	call  ip_showinfo
	ENDC 
; On entry to the protocol handlers,
; IX and HL have the IP datagram base and size respectively,
; IY and DE have the IP payload base and size respectively;
; the handlers may assume the IP datagram is valid, and corrupt all registers
	ld  a, (ipbuffer+IP_VERIHL)
	and  0x0f
	rla 
	rla 
	ld  d, 0
	ld  e, a
	push  ix
	pop  iy
	add  iy, de	; IY = base of data octets
	push  hl
	sbc  hl, de	; C clear before because of ADD
	ex  de, hl	; DE = # of data octets
	pop  hl
; Dispatch
	ld  a, (ipbuffer+IP_PRT)
	cp  IP_ICMP
	jp  z, icmp
	IF  TCP
	cp  IP_TCP
	jp  z, tcp_
	ENDC 
	IF  UDP
	cp  IP_UDP
	jp  z, udp_
	ENDC 
	ld  a, ICMP_PROT_UNREACH
	jp  icmp_senddestunreach
ip_done:		 
	ret 
	IF  CHECKSUM_IP_IN
	IF  SHOW_ERROR
ip_badchecksum:	 
	ex  de, hl
	call  printstringimm
	db  "IP: bad checksum: &" , 0
	jp  printhex2crlf
	ELSE 
ip_badchecksum	EQU  ip_done
	ENDC 
	ENDC 
	IF  SHOW_ERROR
ip_headertoosmall:	 
	call  printstringimm
	db  "IP: header too small: " , 0
	jp  printdeccrlf
ip_headertoobig:	 
	inc  a
	call  printstringimm
	db  "IP: header bigger than datagram: declared &00" , 0
	call  printhex
	call  printstringimm
	db  " vs. actual &" , 0
	ex  de, hl
	jp  printhex2crlf
ip_notv4:	 
	call  printstringimm
	db  "IP: not v4 but v" , 0
	rrca 
	rrca 
	rrca 
	rrca 
	jp  printdeccrlf
ip_fragmented:	 
	call  printstringimm
	db  "IP: fragmented datagram received" , CR, LF, 0
	ret 
ip_sizemismatch:	 
	call  printstringimm
	db  "IP: datagram size mismatch: actual &" , 0
	call  printhex2
	call  printstringimm
	db  " vs declared &" , 0
	add  hl, de
	jp  printhex2crlf
	ELSE 
ip_headertoosmall	EQU  ip_done
ip_headertoobig	EQU  ip_done
ip_notv4	EQU  ip_done
ip_fragmented	EQU  ip_done
ip_sizemismatch	EQU  ip_done
	ENDC 
	IF  SHOW_INFO
ip_showinfo:	 
	push  hl
	IF  SHOW_RAW_IN
; That's already printed a blank line
	ELSE 
	call  printcrlf
	ENDC 
	call  printstringimm
	db  "IP: v4 datagram, size &" , 0
	call  printhex2
	call  printstringimm
	db  ", protocol " , 0
	ld  a, (ipbuffer+IP_PRT)
	call  printdec
	call  printstringimm
	db  ", TOS " , 0
	ld  a, (ipbuffer+IP_TOS)
	call  printdec
	call  printstringimm
	db  ", TTL " , 0
	ld  a, (ipbuffer+IP_TTL)
	call  printdec
	ld  a, (ipbuffer+IP_VERIHL)
	and  0x0f
	sub  5
	jr  z, ip_shownooptions
	rla 
	rla 
	call  printstringimm
	db  ", options size &" , 0
	call  printhex
ip_shownooptions:	 
	call  printstringimm
	db  CR, LF, "IP: remote addr: " , 0
	ld  hl, ipbuffer+IP_SRC3
	call  printipaddr
	call  printstringimm
	db  CR, LF, "IP: local addr:	  " , 0
	ld  hl, ipbuffer+IP_DST3
	call  printipaddr
	call  printcrlf
	pop  hl
	ret 
	ENDC 
ip_swapsrcdest:	 
; Swap source and destination addresses for IP header pointed at by IX
; DE corrupted
	ld  d, (ix+IP_SRC3)
	ld  e, (ix+IP_DST3)
	ld  (ix+IP_SRC3), e
	ld  (ix+IP_DST3), d
	ld  d, (ix+IP_SRC2)
	ld  e, (ix+IP_DST2)
	ld  (ix+IP_SRC2), e
	ld  (ix+IP_DST2), d
	ld  d, (ix+IP_SRC1)
	ld  e, (ix+IP_DST1)
	ld  (ix+IP_SRC1), e
	ld  (ix+IP_DST1), d
	ld  d, (ix+IP_SRC0)
	ld  e, (ix+IP_DST0)
	ld  (ix+IP_SRC0), e
	ld  (ix+IP_DST0), d
	ret 
ip_getID:	 
; Get next value for the IP ID field
;
; On exit:
;	HL = identification number
;
; The ID field is incrememnted on every request, thereby
; ensuring uniqueness within &10000 requests, and so the MSL.
	ld  hl, (ip_ID)
	inc  hl
	ld  (ip_ID), hl
	ret 
ip_ID:		 
	dw  0
ip_getranduserport:	 
; Get "random" port in the user range (>= 1024)
;
; On exit:
;	HL = byte-swapped port number
;
; The port number is unique with approximately second granularity
; and with approximately 8 hours' timespan
;
; The return value is byte-swapped so it can be written little-endianly
	push  af
	push  de
	call  KL_TIME_PLEASE	; DEHL = time in 1/300 s
	ld  l, e	; LH = time in 256 / 300 s
	ld  a, l	; At least 1024?
	and  0xfc
	jr  nz, ip_getranduserport_ok
	set  7, l	; No?  Then set b15
ip_getranduserport_ok:	 
	pop  de
	pop  af
	ret 
ipbufferinsert:	 
	dw  ipbuffer
