; CPC/IP v0.20
; Copyright (c) 1999-2001 Mark RISON, mrison@hotmail.com
*Include config.as 
*List OFF 
; Major FIXMEs:
; - understand DNS better and get resolv right for gethostbyname (incl. CNAMEs)
; - enhance portability (especially re file I/O)
; - rewrite tftpd (support RRQs, use stdio, etc.)
; - Add some token TIMEWAIT state to TCP
; Less major FIXMEs:
; - rename dns_ to dnsd_
; - rename tftp_ to tftpd_
; - make PPP a bit more robust (incl. timeouts and a bit of state)
; - improve initialisation (e.g. zero-init datalink stuff)
; - use file handles for portability
; - sort out authoritativeness in DNSD
; - give the serial driver a chance to try to empty its tx buffers on finalise
; - make tipi's error handling a bit more user-friendly
; - use UCP_OFFSETOF_*
; - redefine TFTP/DNS offsets as being from after UDP header
; - set IP ID field properly
; - set IP TOS field properly
; - handle RRQs in TFTPD
; - fix ICMP stuff
; - put RFC references in each appropriate .z file
main:		 
	call  stdio_init
	call  printstringimm
	db  "CPC/IP v0.20 on " , 0
	ld  hl, config_hostaddr
	call  printipaddr
	call  printstringimm
	IF  IS128K
	db  " (128"
	ELSE 
	db  " (64"
	ENDC 
	db  "K memory configuration, "
	IF  PPP
	db  "PPP"
	ELSE 
	db  "SLIP"
	ENDC 
	db  " link)" , CR, LF
	db  "Copyright (c) 1999-2001 Mark RISON, mrison@hotmail.com" , CR, LF, 0
	call  TXT_CUR_ON
	call  serial_init
	IF  TIMEOUT
	call  timeout_init
	ENDC 
; Initialise servers and clients here
	IF  TFTPD
	call  tftp_init
	ENDC 
	IF  HTTPD
	call  httpd_init
	ENDC 
	IF  DNSD
	call  dns_init
	ENDC 
	IF  TELNET
	call  telnet_init
	ENDC 
	IF  FINGER
	call  finger_init
	ENDC 
	IF  HOST
	call  host_init
	ENDC 
	IF  PING
	call  ping_init
	ENDC 
	IF  IFCONFIG
	call  ifconfig_init
	ENDC 
	IF  RESOLV
	call  resolv_init
	ENDC 
	ld  hl, datalink_flags
	call  datalink_isopen
	call  nz, keyb_displayprompt
main_loop:	 
; Handle timeouts
	IF  TIMEOUT
	call  timeout_dispatch
	ENDC 
; Handle loopback
	IF  LOOPBACK
	call  main_handleloopback
	ENDC 
; Handle serial rx char
	call  serial_getchar
	call  c, main_handlerxchar
; Handle keyboard char
	call  KM_READ_CHAR
	call  c, main_handletxchar
	jr  main_loop
main_done:	 
	pop  hl		; Pop return address for main_handletxchar
; Finalise servers and clients here
	IF  RESOLV
	call  resolv_final
	ENDC 
	IF  IFCONFIG
	call  ifconfig_final
	ENDC 
	IF  PING
	call  ping_final
	ENDC 
	IF  HOST
	call  host_final
	ENDC 
	IF  FINGER
	call  finger_final
	ENDC 
	IF  TELNET
	call  telnet_final
	ENDC 
	IF  DNSD
	call  dns_final
	ENDC 
	IF  HTTPD
	call  httpd_final
	ENDC 
	IF  TFTPD
	call  tftp_final
	ENDC 
	IF  TIMEOUT
	call  timeout_final
	ENDC 
	call  serial_final
	jp  stdio_final
	IF  LOOPBACK
main_handleloopback:	 
; If ipbuffer empty, try to suck packets out of loopback
	call  datalink_isempty
	ret  c
main_getlooped:	 
	call  loopback_remove
	ret  nc
	jr  main_getlooped
	ENDC 
main_handlerxchar:	 
; Pass serial rx chars to datalink if up
	ld  hl, datalink_flags
	call  datalink_isup
	jp  nz, datalink_handlechar
	IF  PPP
; PPP comes up when PPP_ADDR is received
	cp  PPP_ADDR
	ELSE 
; SLIP comes up when SLIP_END is received
	cp  SLIP_END
	ENDC 
	jp  z, datalink_up
; Otherwise print out the received char
	ld  hl, main_parseC
	ld  c, (hl)
	call  util_parsechar
	ld  (hl), c
	ret  c
	cp  NL
	jp  z, printcrlf
	jp  printcharsafe
main_parseC:	 
	db  0
main_handletxchar:	 
; Check for exit character
	cp  KQUIT
	jr  z, main_done
; Check for datalink toggling character
	ld  hl, datalink_flags
	cp  KTOGGLE
	jp  z, datalink_toggle
; Pass to serial tx if datalink down
	call  datalink_isup
	jp  z, serial_putchar
; Drop if datalink not open
	call  datalink_isopen
	ret  z
; Pass to key handler if client active
	ld  hl, (keyb_client_keyhandler)
	ld  b, a
	ld  a, h
	or  l
	ld  a, b
	jp  nz, callhl
; Otherwise accumulate in keyboard line buffer
	cp  CTRL_E
	jr  z, keyb_toggle_echo
	cp  128		; Any top-bit-set char is EOL
	jr  nc, keyb_complete
	cp  32		; Any control char is EOL
	jr  c, keyb_complete
	ld  hl, (keybbufferinsert)
	cp  DEL
	jr  z, keyb_del
	ld  (hl), a
	inc  l
	jr  z, keyb_overflow
	ld  (keybbufferinsert), hl
	ld  hl, keyb_echo
	bit  0, (hl)
	ret  z
	jp  printcharverysafe
keyb_overflow:	 
	ld  a, BEL
	jp  printcharsafe
keyb_toggle_echo:	 
	ld  a, (keyb_echo)
	xor  TRUE
	ld  (keyb_echo), a
	ret 
keyb_del:	 
	dec  l
	inc  l
	jr  z, keyb_overflow
	dec  l
	ld  (keybbufferinsert), hl
	ld  hl, keyb_echo
	bit  0, (hl)
	ret  z
	call  printstringimm
	db  BS, " " , BS, 0
	ret 
keyb_continue:	 
; On entry:
;	C = length
; On exit:
;	A corrupted, F preserved
	call  printstringimm
	db  BEL, 0
	ld  a, c
	ld  (keybbufferinsert), a
	ret 
keyb_complete:	 
	ld  hl, (keybbufferinsert)
	ld  c, l
	ld  (hl), NUL_	; NUL-terminated
	ld  l, 0
	ld  (keybbufferinsert), hl
	ex  de, hl
	ld  hl, (keyb_client_linehandler)
	ld  b, a
	ld  a, h
	or  l
	ld  a, b
	ld  b, 0
	jp  nz, callhl
	cp  ESC		; Accept ESC and CTRL-C as line abort
	jp  z, keyb_displayprompt
	cp  CTRL_C
	jp  z, keyb_displayprompt
	cp  CR		; Only accept CR as valid EOL
	jr  nz, keyb_continue	; Everything else is noise
	call  printcrlf
	ex  de, hl
	call  util_skipspace
	ld  a, (hl)	; Skip empty lines
	or  a
	jr  z, keyb_displayprompt
	ld  c, l
	IF  FINGER
	ld  de, keyb_finger_cmd
	ld  b, keyb_finger_cmdend-keyb_finger_cmd
	call  memicmp
	jp  z, finger_do
	ENDC 
	IF  TELNET
	ld  l, c
	ld  de, keyb_telnet_cmd
	ld  b, keyb_telnet_cmdend-keyb_telnet_cmd
	call  memicmp
	jp  z, telnet_du
	ENDC 
	IF  HOST
	ld  l, c
	ld  de, keyb_host_cmd
	ld  b, keyb_host_cmdend-keyb_host_cmd
	call  memicmp
	jp  z, host_do
	ENDC 
	IF  PING
	ld  l, c
	ld  de, keyb_ping_cmd
	ld  b, keyb_ping_cmdend-keyb_ping_cmd
	call  memicmp
	jp  z, ping_do
	ENDC 
	IF  IFCONFIG
	ld  l, c
	ld  de, keyb_ifconfig_cmd
	ld  b, keyb_ifconfig_cmdend-keyb_ifconfig_cmd
	call  memicmp
	jp  z, ifconfig_do
	ENDC 
	IF  MARK_HACKS
	ld  l, c
	ld  de, keyb_snarf_cmd
	ld  b, keyb_snarf_cmdend-keyb_snarf_cmd
	call  memicmp
	jr  z, snarf_do
	ENDC 
	ld  l, c
	ld  de, keyb_quit_cmd
	ld  b, keyb_quit_cmdend-keyb_quit_cmd
	call  memicmp
	jp  z, main_done
	call  printstringimm
	db  "Bad command" , 0
	jr  keyb_displayprompt
	IF  MARK_HACKS
snarf_do:	 
	call  util_skipspace
	ld  d, h
	ld  e, l
	call  util_skiptospace
	or  a
	sbc  hl, de
	jr  z, keyb_client_done
	ld  b, l
	ex  de, hl
	ld  de, amsdoswritebuf
	call  CAS_OUT_OPEN
	ld  hl, ipbuffer
	ld  de, 1024
	ld  bc, 0
	ld  a, 2
CAS_OUT_DIRECT	EQU  0xbc98
	call  CAS_OUT_DIRECT
	call  CAS_OUT_CLOSE
	call  CAS_OUT_ABANDON
	jr  keyb_client_done
	ENDC 
keyb_client_doing:	 
; On entry to handler:
;	A = char
; On exit from handler:
;	AF, BC, DE, HL, IX, IY corrupted
	ld  (keyb_client_keyhandler), hl
	ret 
keyb_line_doing:	 
; On entry to handler:
;	A = EOL char
;	B = 0
;	C = length (excluding EOL char and NUL)
;	DE -> data (NUL-terminated)
; On exit from handler:
;	AF, BC, DE, HL, IX, IY corrupted
	ld  (keyb_client_linehandler), hl
	push  hl
	ld  hl, 0
	ld  (keyb_client_keyhandler), hl
	pop  hl
	ret 
keyb_client_done:	 
	ld  hl, 0
	ld  (keyb_client_keyhandler), hl
	ld  (keyb_client_linehandler), hl
keyb_displayprompt:	 
	call  printcrlfmaybe
	call  printstringimm
	db  PROMPT, " " , 0
	ret 
	IF  TELNET
keyb_telnet_cmd:	 
	db  "telnet"
keyb_telnet_cmdend:	 
	ENDC 
	IF  FINGER
keyb_finger_cmd:	 
	db  "finger"
keyb_finger_cmdend:	 
	ENDC 
	IF  HOST
keyb_host_cmd:	 
	db  "host"
keyb_host_cmdend:	 
	ENDC 
	IF  PING
keyb_ping_cmd:	 
	db  "ping"
keyb_ping_cmdend:	 
	ENDC 
	IF  IFCONFIG
keyb_ifconfig_cmd:	 
	db  "ifconfig"
keyb_ifconfig_cmdend:	 
	ENDC 
	IF  MARK_HACKS
keyb_snarf_cmd:	 
	db  "snarf"
keyb_snarf_cmdend:	 
	ENDC 
keyb_quit_cmd:	 
	db  "quit"
keyb_quit_cmdend:	 
keybbufferinsert:	 
	dw  keybbuffer
keyb_echo:	 
	db  TRUE
keyb_client_keyhandler:	 
	dw  0
keyb_client_linehandler:	 
	dw  0
; These two must be below &4000 in the 128K memory configuration
; (i.e. below bank 1 -- which gets switched out!)
	IF  INTDRIVEN
*Include serialdd.as 
	ENDC 
	IF  LOOPBACK
*Include loopback.as 
	ENDC 
; The following can be anywhere, fortunately!
	IF  PPP
PPP_FCSTAB16_PAGEALIGNED	EQU  0
*Include fcstab16.as 
*Include ppp.as 
	ELSE 
*Include slip.as 
	ENDC 
*Include stdio.as 
	IF  TIMEOUT
*Include timeout.as 
	ENDC 
*Include ip.as 
*Include cs.as 
*Include icmp.as 
	IF  TCP
*Include tcp.as 
	ENDC 
	IF  UDP
*Include udp.as 
	ENDC 
; The following must be above &3fff
; (i.e. above bank 0 -- where event blocks can't be!)
	IF  TFTPD
*Include tftpd.as 
	ENDC 
	IF  HTTPD
*Include httpd.as 
	ENDC 
	IF  DNSD
*Include dnsd.as 
	ELSE 
dns_resourcerecords	EQU  0
	ENDC 
	IF  TELNET
*Include telnet.as 
*Include cons.as 
	ENDC 
	IF  FINGER
*Include finger.as 
	ENDC 
	IF  HOST
*Include host.as 
	ENDC 
	IF  PING
*Include ping.as 
	ENDC 
	IF  IFCONFIG
*Include ifconfig.as 
	ENDC 
	IF  RESOLV
*Include resolv.as 
	ENDC 
	IF  HOSTS
*Include hosts.as 
	ELSE 
hosts_txt	EQU  0
	ENDC 
*Include printlib.as 
*Include utillib.as 
	IF  UDP .or. TCP
*Include linklist.as 
	ENDC 
; The following must be above &7fff in the 128K memory configuration
; (i.e. bank 2 -- where both event blocks and banking code can be!)
	IF  POLLED
	IF  IS128K
	ORG  specialcode
	ENDC 
*Include serialdd.as 
	ENDC 
*List ON 
	end 
