; CPC/IP serial device driver for Amstrad/Pace RS232C
; and CPC Amstrad International (1992-06/07) serial interfaces
; Copyright (c) 1999-2001 Mark RISON
; A serial device driver must export the following routines
; (the entry/exit conditions are described here at the top of each):
; serial_init, serial_final, serial_getchar, serial_putchar
; A serial device driver may expect the following imported symbols:
; BIT_RATE,
; PARITY_NONE, PARITY_ODD, PARITY_EVEN,
; STOP_1, STOP_15, STOP_2,
; MEMCONFIG, IS64K, IS128K,
; INTDRIVEN, POLLED,
; HWFLOWCONTROL,
; serialtxbuffer, serialtxbufsiz, serialrxbuffer, serialrxbufsiz
; Note the Pace interface does not support being interrupt-driven
; unless a hardware modification has been made.
; The CPCAI interface does not support being interrupt-driven,
; nor does it support datarates other than 2400 bps, and it
; mandates hardware flow control.
	IF  AMSTRAD .or. PACE
SIO_A_DATA	EQU  0xfadc
SIO_A_CONTROL	EQU  0xfadd
HBIT_RATE	EQU  BIT_RATE/100
INV_RATE	EQU  1250/HBIT_RATE
	IF  PARITY_NONE
PARITY_BITS	EQU  0
	ENDC 
	IF  PARITY_ODD
PARITY_BITS	EQU  1
	ENDC 
	IF  PARITY_EVEN
PARITY_BITS	EQU  3
	ENDC 
	IF  STOP_1
STOP_BITS	EQU  4
	ENDC 
	IF  STOP_1_5
STOP_BITS	EQU  8
	ENDC 
	IF  STOP_2
STOP_BITS	EQU  12
	ENDC 
WR0_RESETTXINT	EQU  0x28
WR0_RESET	EQU  0x18	; channel reset
	IF  INTDRIVEN
WR1_INIT	EQU  0x1a	; tx and all rx int
	ELSE 
WR1_INIT	EQU  0x00	; no ints
	ENDC 
	IF  HWFLOWCONTROL
	IF  PACE
	error  "Pace serial interface doesn't support hardware flow control"
; If the hardware modification has been made,
; it should be considered an honorary Amstrad serial interface!
	ENDC 
WR3_INIT	EQU  0xe1	; 8 data rx, hw flow, rx enable
	ELSE 
WR3_INIT	EQU  0xc1	; 8 data rx, no hw flow, rx enable
	ENDC 
WR4_INIT	EQU  0x40 .or. STOP_BITS .or. PARITY_BITS	; x16 clk
WR5_INIT	EQU  0xea	; DTR+RTS active, 8 data tx, tx enable
WR5_RTSACTIVE	EQU  0xea
WR5_RTSINACTIVE	EQU  0xe8
WR1		EQU  1
WR3		EQU  3
WR4		EQU  4
WR5		EQU  5
RR0_RESETTEST_MASK	EQU  0x07
RR0_RESETTEST_RESULT	EQU  0x04
CTC_CT0		EQU  0xfbdc
CTC_CT1		EQU  0xfbdd
CTC_MODE	EQU  0xfbdf
MODE_CT0	EQU  0x36
MODE_CT1	EQU  0x76
	ELSE 		; CPCAI
MC6850_CONTROL	EQU  0xf8dc
MC6850_DATA	EQU  0xf8dd
	IF  HWFLOWCONTROL
	ELSE 
	error  "CPCAI serial device driver mandates use of hardware flow control"
	ENDC 
	IF  INTDRIVEN
	error  "CPCAI serial device driver only supports polling"
	ENDC 
	IF  STOP_1_5
	error  "CPCAI serial device driver doesn't support 1 1/2 stop bits"
	ENDC 
	IF  STOP_2
	IF  PARITY_EVEN .or. PARITY_ODD
	error  "CPCAI serial device driver doesn't support 2 stop bits with parity"
	ENDC 
FRAMING_BITS	EQU  0x10
	ELSE 		; !STOP_2
	IF  PARITY_NONE
FRAMING_BITS	EQU  0x14
	ENDC 
	IF  PARITY_EVEN
FRAMING_BITS	EQU  0x18
	ENDC 
	IF  PARITY_ODD
FRAMING_BITS	EQU  0x1c
	ENDC 
	ENDC 		; STOP_2
	IF  BIT_RATE-2400
	error  "CPCAI serial device driver only supports 2400 bps"
	ELSE 
DIVIDE_16	EQU  1
MASTER_RESET	EQU  3
	ENDC 
RTSASSERTED_TXINTOFF	EQU  0x00
RTSASSERTED_TXINTON	EQU  0x20
RTSDEASSERTED_TXINTOFF	EQU  0x40
RTSASSERTED_TXBRK_TXINTOFF	EQU  0x60
RXINTOFF	EQU  0x00
RXINTON		EQU  0x80
MC6850_RESETTEST_MASK	EQU  0xf3
MC6850_RESETTEST_RESULT	EQU  0x02
MC6850_CONTROL_RTSASSERTED	EQU  DIVIDE_16 .or. FRAMING_BITS .or. RTSASSERTED_TXINTOFF .or. RXINTOFF
MC6850_CONTROL_RTSDEASSERTED	EQU  DIVIDE_16 .or. FRAMING_BITS .or. RTSDEASSERTED_TXINTOFF .or. RXINTOFF
	ENDC 
TXACTIVE	EQU  0
	IF  IS128K
BANK_TXBUFFER	EQU  0x7fc5	; bank 5 at &4000
BANK_RXBUFFER	EQU  0x7fc4	; bank 4 at &4000
BANK_RESTORE	EQU  0x7fc0	; bank 1 at &4000
	ENDC 
serial_init:	 
; Initialise serial device and driver
; AF, BC, DE, HL may be corrupted
; Interrupts may be disabled and then reenabled
	di 
	IF  INTDRIVEN
INTERRUPT_ENTRY	EQU  0x0038
EXT_INTERRUPT	EQU  0x003b
NEW_CODE	EQU  0xbdc4	; 6 unused bytes
; Allow Z80 support chip interrupts to work
; (based on Soft 968 App. XIII.7, but optimised)
	ld  hl, (INTERRUPT_ENTRY+1)	; HL = ADDRESS_X
	ld  de, 5
	add  hl, de	; HL = ADDRESS_X + 5
	ld  de, call_new_code	; Copy the CALL NEW_CODE patch
	ex  de, hl	; to ADDRESS_X + 5
	ld  bc, 4
	ldir 
	ld  hl, new_code_patch	; Copy the NEW_CODE patch itself
	ld  de, NEW_CODE	; to NEW_CODE
	ld  bc, 6
	ldir 
; Patch in external interrupt handler
; Assume sole external interrupt source
	ld  hl, ext_interrupt_patch
	ld  de, EXT_INTERRUPT
	ld  bc, 3
	ldir 
	ENDC 		; INTDRIVEN
	IF  AMSTRAD .or. PACE
; Initialise CTC
	ld  bc, CTC_MODE
	ld  a, MODE_CT0
	out  (c), a
	ld  c, CTC_CT0 .and. 0xff
	ld  a, INV_RATE .and. 0xff
	out  (c), a
	ld  a, INV_RATE/256
	out  (c), a
	ld  c, CTC_MODE .and. 0xff
	ld  a, MODE_CT1
	out  (c), a
	ld  c, CTC_CT1 .and. 0xff
	ld  a, INV_RATE .and. 0xff
	out  (c), a
	ld  a, INV_RATE/256
	out  (c), a
; Initialise SIO
	ld  bc, SIO_A_CONTROL
	ld  d, WR0_RESET
	out  (c), d
	in  a, (c)
	ld  d, WR4	; Technical manual says do WR4 before 1/3/5
	out  (c), d
	ld  d, WR4_INIT
	out  (c), d
	ld  d, WR3
	out  (c), d
	ld  d, WR3_INIT
	out  (c), d
	ld  d, WR5
	out  (c), d
	ld  d, WR5_INIT
	out  (c), d
	ld  d, WR1	; Appnote suggests doing WR1 last
	out  (c), d
	ld  d, WR1_INIT
	out  (c), d
; Check it's all (apparently) there!
	and  RR0_RESETTEST_MASK
	cp  RR0_RESETTEST_RESULT
	ELSE 		; CPCAI
; Initialise MC6850
	ld  bc, MC6850_CONTROL
	ld  a, MASTER_RESET .or. FRAMING_BITS .or. RTSDEASSERTED_TXINTOFF .or. RXINTOFF
	out  (c), a
	ld  a, MC6850_CONTROL_RTSASSERTED
	out  (c), a
; Check it's all (apparently) there!
	in  a, (c)
	and  MC6850_RESETTEST_MASK
	cp  MC6850_RESETTEST_RESULT
	ENDC 
	ei 
	jr  nz, serial_notresponding
	IF  POLLED
; Initialise fast ticker events for polling
KL_NEW_FAST_TICKER	EQU  0xbce0
	ld  hl, serial_fasttickerblock
	ld  de, serial_interrupt_handler
	ld  bc, 0xc100	; Express asynchronous event at near address
	call  KL_NEW_FAST_TICKER
	ENDC 		; POLLED
	call  printstringimm
	db  "Serial interface driver initialised (" , 0
	ld  hl, BIT_RATE
	call  printdec2
	call  printstringimm
	db  " bps, 8"
	IF  PARITY_NONE
	db  "N"
	ENDC 
	IF  PARITY_EVEN
	db  "E"
	ENDC 
	IF  PARITY_ODD
	db  "O"
	ENDC 
	IF  STOP_1
	db  "1"
	ENDC 
	IF  STOP_1_5
	db  "1 1/2"
	ENDC 
	IF  STOP_2
	db  "2"
	ENDC 
	db  ", "
	IF  HWFLOWCONTROL
	ELSE 
	db  "no "
	ENDC 
	db  "hw flow, "
	IF  INTDRIVEN
	db  "int-driven"
	ELSE 
	db  "polled"
	ENDC 
	db  ")" , CR, LF, 0
	ret 
serial_notresponding:	 
	call  printstringimm
	IF  AMSTRAD
	db  "*** Amstrad serial interface not responding ***" , CR, LF, 0
	ELSE 
	IF  PACE
	db  "*** Pace serial interface not responding ***" , CR, LF, 0
	ELSE 
	db  "*** CPCAI serial interface not responding ***" , CR, LF, 0
	ENDC 
	ENDC 
	ret 
	IF  POLLED
serial_fasttickerblock:	 
	ds  9
	ENDC 
serial_final:	 
; Close down serial device
; AF, BC, DE, HL may be corrupted
; Interrupts may be disabled and then reenabled
	IF  POLLED
KL_DEL_FAST_TICKER	EQU  0xbce6
	ld  hl, serial_fasttickerblock
	call  KL_DEL_FAST_TICKER
	ENDC 		; POLLED
	IF  AMSTRAD .or. PACE
	ld  bc, SIO_A_CONTROL
	ld  a, WR0_RESET
	out  (c), a
	ELSE 
	ld  bc, MC6850_CONTROL
	ld  a, MASTER_RESET .or. FRAMING_BITS .or. RTSASSERTED_TXINTOFF .or. RXINTOFF
	out  (c), a
	ENDC 
	ret 
serial_getchar:	 
; On exit, A = next character from the serial rx buffer
; C clear if no character available (A corrupted)
; Other flags corrupted
; Interrupts enabled
	push  de
	push  hl
	or  a
	di 
	ld  hl, (rxbufferinsert)
	ld  de, (rxbufferremove)
	sbc  hl, de	; HL = # chars in buffer (modulo buffer size)
	jr  z, serial_getchar_none	; C clear if no char
	IF  IS128K
	push  bc
	ENDC 
	IF  HWFLOWCONTROL
	xor  a		; Reactivate RTS if # free = &ff
	cp  l
	jr  nz, serial_nortsactivateneeded
	ld  a, h
	cpl 
	and  serialrxbufsiz/256-1
	jr  nz, serial_nortsactivateneeded
	IF  IS64K
	push  bc
	ENDC 
	IF  AMSTRAD .or. PACE
	ld  bc, SIO_A_CONTROL
	ld  a, WR5
	out  (c), a
	ld  a, WR5_RTSACTIVE
	ELSE 
	ld  bc, MC6850_CONTROL
	ld  a, MC6850_CONTROL_RTSASSERTED
	ENDC 
	out  (c), a
	IF  IS64K
	pop  bc
	ENDC 
serial_nortsactivateneeded:	 
	ENDC 		; HWFLOWCONTROL
	IF  IS128K
	ld  bc, BANK_RXBUFFER
	out  (c), c
	ENDC 
	ld  a, (de)
	IF  IS128K
	ld  c, BANK_RESTORE AND 0xff
	out  (c), c
	pop  bc
	ENDC 
	inc  de
	IF  IS128K
	set  6, d
	res  7, d
	ELSE 
	set  2, d
	set  3, d
	res  4, d
	ENDC 
	ld  (rxbufferremove), de
	scf 		; C set if char
serial_getchar_none:	 
	ei 
	pop  hl
	pop  de
	ret 
serial_putchar:	 
; On entry, A = character to put in the serial tx buffer
; All regs preserved
; Interrupts enabled
; The behaviour is undefined (but not fatal) if the buffer is full
	push  af
	IF  IS128K
	push  bc
	ENDC 
	push  hl
	ld  hl, txflags
	di 
	bit  TXACTIVE, (hl)
	jr  nz, serial_putchar_txisactive
	IF  IS64K
	push  bc
	ENDC 
	IF  AMSTRAD .or. PACE
	ld  bc, SIO_A_DATA
	ELSE 
	ld  bc, MC6850_DATA
	ENDC 
	out  (c), a
	IF  IS64K
	pop  bc
	ENDC 
	set  TXACTIVE, (hl)
	jr  serial_putchar_txwasinactive
serial_putchar_txisactive:	 
	ld  hl, (txbufferinsert)
	IF  IS128K
	ld  bc, BANK_TXBUFFER
	out  (c), c
	ENDC 
	ld  (hl), a
	IF  IS128K
	ld  c, BANK_RESTORE AND 0xff
	out  (c), c
	ENDC 
	inc  hl
	IF  IS128K
	set  6, h
	res  7, h
	ELSE 
	res  2, h
	ENDC 
	ld  (txbufferinsert), hl
serial_putchar_txwasinactive:	 
	ei 
	pop  hl
	IF  IS128K
	pop  bc
	ENDC 
	pop  af
	ret 
serial_interrupt_handler:	 
; AF, BC, DE, HL may be corrupted (see Soft 968
; pages 11.2 and 18.21 for interrupt-driven, and
; pages 11.3 and 15.200 for fast ticker-driven)
; If interrupt-driven, assume sole external interrupt source,
; and don't try to cope with rx and tx interrupts simultaneously.
;
; If polled, rx as many characters as there are in the FIFO
; (could be up to 3 for Z80-SIO) and then tx a character if waiting
; (there's no tx FIFO so no point looping).
	IF  AMSTRAD .or. PACE
	ld  bc, SIO_A_CONTROL
	ELSE 
	ld  bc, MC6850_CONTROL
	ENDC 
	in  a, (c)
	rra 		; b0 set if rx available
	jr  nc, serial_txinterrupt
serial_rxinterrupt:	 
	IF  AMSTRAD .or. PACE
	dec  c		; SIO_A_DATA
	ELSE 
	inc  c		; MC6850_DATA
	ENDC 
	in  a, (c)
	ld  hl, (rxbufferremove)
	ld  de, (rxbufferinsert)
	IF  IS128K
	ld  bc, BANK_RXBUFFER
	out  (c), c
	ENDC 
	ld  (de), a
	IF  IS128K
	ld  c, BANK_RESTORE .and. 0xff
	out  (c), c
	ENDC 
	inc  de		; COULD argue rx buffer full should never happen if HWFLOWCONTROL (because of RTS) and so not test (but still need SBC for RTS test)
	IF  IS128K
	set  6, d
	res  7, d
	ELSE 
	set  2, d
	set  3, d
	res  4, d
	ENDC 
	or  a
	sbc  hl, de	; HL = # free in buffer + 1 after this rx buffered
	jr  z, serial_rxbufferfull	; (modulo buffer size)
	ld  (rxbufferinsert), de
	IF  HWFLOWCONTROL
	ld  a, h	; Deactivate RTS if # free + 1 <= &ff
	and  serialrxbufsiz/256-1	; (i.e. # free < &ff)
	jr  z, serial_rtsdeactivateneeded	; (don't bother checking lsb as this should stop rx ints quite quickly)
serial_rxbufferfull:	 		; Should never happen if RTS honoured
	IF  INTDRIVEN
	reti 
	ELSE 
	jr  serial_interrupt_handler
	ENDC 
serial_rtsdeactivateneeded:	 
	IF  AMSTRAD .or. PACE
	ld  bc, SIO_A_CONTROL
	ld  a, WR5
	out  (c), a
	ld  a, WR5_RTSINACTIVE
	ELSE 
	ld  bc, MC6850_CONTROL
	ld  a, MC6850_CONTROL_RTSDEASSERTED
	ENDC 
	out  (c), a
	ELSE 		; !HWFLOWCONTROL
serial_rxbufferfull:	 
	ENDC 		; HWFLOWCONTROL
	IF  INTDRIVEN
	reti 
	ELSE 
	jr  serial_interrupt_handler
	ENDC 
serial_txinterrupt:	 
	IF  POLLED
	IF  AMSTRAD .or. PACE
	rra 		; b2 set if tx empty
	rra 
	ELSE 
	rra 		; b1 set if tx empty
	ENDC 
	jr  nc, serial_interrupt_done
	ENDC 
	ld  hl, (txbufferinsert)
	ld  de, (txbufferremove)
	xor  a
	IF  POLLED
	 		; Fix firmware bug for express async events
	ld  (serial_fasttickerblock+4), a	; (see Soft 968 App. XIII.3)
	ENDC 
	sbc  hl, de
	jr  z, serial_nothingmoretosend
	IF  IS128K
	ld  bc, BANK_TXBUFFER
	out  (c), c
	ENDC 
	ld  a, (de)
	IF  IS128K
	ld  c, BANK_RESTORE .and. 0xff
	out  (c), c
	ENDC 
	IF  AMSTRAD .or. PACE
	ld  bc, SIO_A_DATA
	ELSE 
	ld  bc, MC6850_DATA
	ENDC 
	out  (c), a
	inc  de
	IF  IS128K
	set  6, d
	res  7, d
	ELSE 
	res  2, d
	ENDC 
	ld  (txbufferremove), de
	reti 
serial_nothingmoretosend:	 
	ld  (txflags), a	; Assumes txflags only contains TXACTIVE
	IF  AMSTRAD .or. PACE
	ld  a, WR0_RESETTXINT
	out  (c), a
	ENDC 
	reti 
	IF  POLLED
serial_interrupt_done:	 
	xor  a		; Fix firmware bug for express async events
	ld  (serial_fasttickerblock+4), a	; (see Soft 968 App. XIII.3)
	reti 
	ENDC 
	IF  INTDRIVEN
ext_interrupt_patch:	 
	jp  serial_interrupt_handler
call_new_code:	 
	call  NEW_CODE
	nop 
new_code_patch:	 
	ld  a, c
	scf 
	ex  af, af'
	ei 
	reti 
	ENDC 		; INTDRIVEN
rxbufferinsert:	 
	dw  serialrxbuffer
rxbufferremove:	 
	dw  serialrxbuffer
txbufferinsert:	 
	dw  serialtxbuffer
txbufferremove:	 
	dw  serialtxbuffer
txflags:		 
	db  0
