; CPC/IP SLIP layer
; Copyright (c) 1999-2000 Mark RISON
; See RFC1055
SLIP_END	EQU  0xc0
SLIP_ESC	EQU  0xdb
SLIP_ESC_END	EQU  0xdc
SLIP_ESC_ESC	EQU  0xdd
datalink_flags:	 
	db  0
SLIP_OPEN_BIT	EQU  7
SLIP_INESC_BIT	EQU  6
datalink_isup:	 		; There's no distinction in SLIP,
datalink_isopen:	 		; since there's no establishment phase
	bit  SLIP_OPEN_BIT, (hl)
	ret 
datalink_up:	 
	set  SLIP_OPEN_BIT, (hl)
	call  printcrlfmaybe
	call  printstringimm
	db  "SLIP: link established" , 0
	jp  keyb_displayprompt
datalink_toggle:	 
	bit  SLIP_OPEN_BIT, (hl)
	jr  z, datalink_up
	res  SLIP_OPEN_BIT, (hl)
	call  printcrlfmaybe
	call  printstringimm
	db  "SLIP: link terminated" , CR, LF, 0
	ret 
datalink_isempty:	 
	ld  hl, (ipbufferinsert)
	ld  de, -ipbuffer-1
	add  hl, de	; Carry clear iff (ipbufferinsert) at ipbuffer
	ret 
datalink_handlechar:	 
	bit  SLIP_INESC_BIT, (hl)
	jr  nz, slip_inesc
	cp  SLIP_END
	jr  z, slip_eop
	cp  SLIP_ESC
	jr  z, slip_intoesc
slip_wasslipesc:	 
	ld  hl, (ipbufferinsert)
	ld  (hl), a
	inc  hl
	bit  2, h
	ret  nz
	ld  (ipbufferinsert), hl
	ret 
slip_inesc:	 
	res  SLIP_INESC_BIT, (hl)
	cp  SLIP_ESC_ESC
	jr  nz, slip_notslipescesc
	ld  a, SLIP_ESC
slip_notslipescesc:	 
	cp  SLIP_ESC_END	; Anything other than SLIP_ESC_ESC and _END
	jr  nz, slip_wasslipesc	; is left untouched, as per RFC1055
	ld  a, SLIP_END
	jr  slip_wasslipesc
slip_intoesc:	 
	set  SLIP_INESC_BIT, (hl)
	ret 
slip_eop:	 
; It's looking good!
	IF  SHOW_RAW_IN
	ld  hl, (ipbufferinsert)
	ld  de, ipbuffer
	or  a
	sbc  hl, de
	ret  z		; Always quietly dump zero-len
	call  printstringimm
	db  CR, LF, "[ " , 0
slip_printdatagram:	 
	ld  a, (de)
	call  printhex
	call  printstringimm
	db  " " , 0
	inc  de
	dec  hl
	ld  a, h
	or  l
	jr  nz, slip_printdatagram
	call  printstringimm
	db  "]" , CR, LF, 0
	ENDC 
	call  ip_handle
datalink_sop:	 
	ld  hl, ipbuffer
	ld  (ipbufferinsert), hl
	ret 
datalink_sendpacket:	 
slip_sendpacket:	 
; Send packet of size BC with base IX
; AF, BC, HL corrupted
	IF  DROP_OUT
	ld  a, (slip_send_drop)
	add  a, 256/DROP_OUT
	ld  (slip_send_drop), a
	ret  c
	ENDC 
; Check for loopback
	push  ix
	pop  hl
	IF  LOOPBACK
	ld  a, (ix+IP_DST3)
	cp  LOOPBACK_ADDR
	jp  z, loopback_insert
	ENDC 
	ld  a, SLIP_END
	call  serial_putchar
	IF  SHOW_RAW_OUT
	call  printstringimm
	db  "{" , 0
	ENDC 
slip_sendpacket_more:	 
	ld  a, (hl)
	IF  SHOW_RAW_OUT
	call  printstringimm
	db  " " , 0
	call  printhex
	ENDC 
	cp  SLIP_END
	jr  nz, slip_sendpacket_notend
	ld  a, SLIP_ESC
	call  serial_putchar
	ld  a, SLIP_ESC_END
slip_sendpacket_notend:	 
	cp  SLIP_ESC
	jr  nz, slip_sendpacket_notesc
	ld  a, SLIP_ESC
	call  serial_putchar
	ld  a, SLIP_ESC_ESC
slip_sendpacket_notesc:	 
	call  serial_putchar
	inc  hl
	dec  bc
	ld  a, b
	or  c
	jr  nz, slip_sendpacket_more
	ld  a, SLIP_END
	call  serial_putchar
	IF  SHOW_RAW_OUT
	call  printstringimm
	db  " }" , CR, LF, 0
	ENDC 
	ret 
	IF  DROP_OUT
slip_send_drop:	 
	db  0
	ENDC 
