; CPC/IP UDP layer
; Copyright (c) 1999-2001 Mark RISON
; The UDP layer exports the following routines
; (the entry/exit conditions are described here at the top of each):
; udp_bind, udp_close
UDP_SRC1	EQU  0
UDP_SRC0	EQU  1
UDP_DST1	EQU  2
UDP_DST0	EQU  3
UDP_LEN1	EQU  4
UDP_LEN0	EQU  5
UDP_CSM1	EQU  6
UDP_CSM0	EQU  7
UDP_HDRSIZE	EQU  8
udp_:		 
; The minimal UDP datagram size is 8
	push  hl
	ld  hl, -UDP_HDRSIZE
	add  hl, de
	pop  hl
	jp  nc, udp_toosmall
; Does the received datagram size match the declared size?
	push  hl
	ld  h, (iy+UDP_LEN1)
	ld  l, (iy+UDP_LEN0)
	or  a
	sbc  hl, de
	pop  hl
	jp  nz, udp_sizemismatch
; Is there a checksum?  If so, check it
	ld  a, (iy+UDP_CSM1)
	or  (iy+UDP_CSM0)
	push  de
	call  nz, cs_calcudpchecksum
	jp  nz, udp_badchecksum
	pop  de
; It's looking good!
	IF  SHOW_INFO
	call  udp_showinfo
	ENDC 
; Go through UDP bind list looking for match
	push  hl
	ld  hl, (udp_client_head)
	call  ucp_dispatch
	pop  hl
; Too bad...
	IF  SHOW_ERROR
	call  printstringimm
	db  "UDP: datagram on unbound local port " , 0
	ld  h, (iy+UDP_DST1)
	ld  l, (iy+UDP_DST0)
	call  printdec2
	call  printstringimm
	db  " (or remote " , 0
	push  ix
	pop  hl
	ld  bc, IP_SRC3
	add  hl, bc
	call  printipaddr
	call  printstringimm
	db  ":" , 0
	ld  h, (iy+UDP_SRC1)
	ld  l, (iy+UDP_SRC0)
	call  printdec2
	call  printstringimm
	db  " rejected)" , CR, LF, 0
	ENDC 
	ld  a, ICMP_PORT_UNREACH
	call  icmp_senddestunreach
udp_done:	 
	ret 
udp_client_head:	 
	dw  0
udp_bind:	 
; UDP bind
;
; On entry:
;	HL -> bind structure
;
; On exit:
;	All regs preserved
;
; Note no check is made that the given bind structure isn't already
; on the bind list -- ensure this doesn't happen!
;
; The bind structure is:
;	word	next		(system-maintained; don't ever write to it)
;	dword	remaddr		(big-endian)
;	word	remport		(big-endian)
;	word	locport		(big-endian)
;	word	handler		(little-endian)
;
; The first octet of the remaddr and the two octets of the remport
; may independently be set to all-zeroes to mean "any".
; The octets of the remaddr and the remport may all be set to
; all-ones to mean "none" (effectively temporarily suspending the UDP bind).
	push  ix
	push  de
	ld  ix, udp_client_head
	call  ll_insert
	pop  de
	pop  ix
	ret 
udp_close:	 
; UDP close
;
; On entry:
;	HL -> bind structure
;
; On exit:
;	All regs preserved
;
; Note if the structure isn't on the bind list, nothing happens.
	push  ix
	ld  ix, udp_client_head
	call  ll_remove
	pop  ix
	ret 
	IF  SHOW_ERROR
udp_badchecksum:	 
	call  printstringimm
	db  "UDP: bad checksum: &" , 0
	pop  hl
	ex  de, hl
	jp  printhex2crlf
udp_toosmall:	 
	call  printstringimm
	db  "UDP: datagram too small: " , 0
	ld  a, e
	call  printdec
	jp  printcrlf
udp_sizemismatch:	 
	call  printstringimm
	db  "UDP: datagram size mismatch: declared &" , 0
	ld  h, (iy+UDP_LEN1)
	ld  l, (iy+UDP_LEN0)
	call  printhex2
	call  printstringimm
	db  " vs. actual &" , 0
	ex  de, hl
	jp  printhex2crlf
	ELSE 
udp_badchecksum	EQU  udp_done
udp_toosmall	EQU  udp_done
udp_sizemismatch	EQU  udp_done
	ENDC 
	IF  SHOW_INFO
udp_showinfo:	 
	push  hl
	call  printstringimm
	db  "UDP: datagram size: &" , 0
	ld  h, (iy+UDP_LEN1)
	ld  l, (iy+UDP_LEN0)
	call  printhex2crlf
	call  printstringimm
	db  "UDP: source port: " , 0
	ld  h, (iy+UDP_SRC1)
	ld  l, (iy+UDP_SRC0)
	call  printdec2
	call  printcrlf
	call  printstringimm
	db  "UDP: dest port:   " , 0
	ld  h, (iy+UDP_DST1)
	ld  l, (iy+UDP_DST0)
	call  printdec2
	call  printcrlf
	pop  hl
	ret 
	ENDC 
